searchData={"items":[{"type":"module","title":"jsonpull","doc":"These functions both read and transform JSON data into Erlang types. The main reading functions return {ok, {Value, Rest}} or {error, Atom} . They sit a level above the jsonpull_read module, so they also skip whitespace.","ref":"jsonpull.html"},{"type":"function","title":"jsonpull.array/1","doc":"Try to read an array from the JSON. If it's there, the atom 'begin_array' will be returned. Structured types are handled differently in pull parsers. For more info, read the Overview.","ref":"jsonpull.html#array/1"},{"type":"function","title":"jsonpull.boolean/1","doc":"Try to read a boolean from the JSON. If it's there, it will be returned as the atom 'true' or 'false' .","ref":"jsonpull.html#boolean/1"},{"type":"function","title":"jsonpull.element/1","doc":"Try to read the next array element from the JSON. If there is an element, it will be represented as the atom 'element' If the array is ending, the atom 'end_array' will be returned. Structured types are handled differently in pull parsers. For more info, read the Overview.","ref":"jsonpull.html#element/1"},{"type":"function","title":"jsonpull.existing_atom/1","doc":"Try to read a string from the JSON as an already existing atom. This function will not parse escape sequences. See also: string/1 .","ref":"jsonpull.html#existing_atom/1"},{"type":"function","title":"jsonpull.float/1","doc":"Try to read a float from the JSON. This function will discriminate against integers and only return floats. See also: integer/1 , number/1 .","ref":"jsonpull.html#float/1"},{"type":"function","title":"jsonpull.integer/1","doc":"Try to read an integer from the JSON. This function will discriminate against floats and only return integers. See also: float/1 , number/1 .","ref":"jsonpull.html#integer/1"},{"type":"function","title":"jsonpull.iolist/1","doc":"Try to read a string from the JSON as an iolist. This is like reading a string, except escape sequences are handled differently. This may be more efficient in some cases. See also: string/1 .","ref":"jsonpull.html#iolist/1"},{"type":"function","title":"jsonpull.key/1","doc":"Try to read the next object key from the JSON. If there is a key, it will be returned as a binary string. If the object is ending, the atom 'end_object' will be returned. Structured types are handled differently in pull parsers. For more info, read the Overview.","ref":"jsonpull.html#key/1"},{"type":"function","title":"jsonpull.null/1","doc":"Try to read a null from the JSON. If it's there, it will be returned as the atom 'null' .","ref":"jsonpull.html#null/1"},{"type":"function","title":"jsonpull.number/1","doc":"Try to read a number from the JSON. If it's there, it will be returned as the original binary string version. See also: float/1 , integer/1 .","ref":"jsonpull.html#number/1"},{"type":"function","title":"jsonpull.object/1","doc":"Try to read an object from the JSON. If there is an object, the atom 'begin_object' will be returned. Structured types are handled differently in pull parsers. For more info, read the Overview.","ref":"jsonpull.html#object/1"},{"type":"function","title":"jsonpull.skip_value/1","doc":"Skips over the next value in the provided JSON binary.","ref":"jsonpull.html#skip_value/1"},{"type":"function","title":"jsonpull.string/1","doc":"Try to read a string from the JSON. If it's there, it will be returned as a binary string with all escape sequences processed. See also: iolist/1 .","ref":"jsonpull.html#string/1"},{"type":"function","title":"jsonpull.strip/1","doc":"Strips the quotes from a raw JSON string. See also: unescape_and_strip/1 .","ref":"jsonpull.html#strip/1"},{"type":"function","title":"jsonpull.unescape_and_strip/1","doc":"Turn a raw JSON string into a properly translated Erlang iolist. Note: You may need to use unicode:characters_to_binary/list afterwards if you don't want an iolist. Using iolist_to_binary will only work if there are no weird unicode escape sequences!","ref":"jsonpull.html#unescape_and_strip/1"},{"type":"module","title":"jsonpull_expect","doc":"These functions will read a value or error out immediately, with useful errorinfo applied on top.","ref":"jsonpull_expect.html"},{"type":"function","title":"jsonpull_expect.array/1","doc":null,"ref":"jsonpull_expect.html#array/1"},{"type":"function","title":"jsonpull_expect.boolean/1","doc":null,"ref":"jsonpull_expect.html#boolean/1"},{"type":"function","title":"jsonpull_expect.element/1","doc":null,"ref":"jsonpull_expect.html#element/1"},{"type":"function","title":"jsonpull_expect.existing_atom/1","doc":null,"ref":"jsonpull_expect.html#existing_atom/1"},{"type":"function","title":"jsonpull_expect.float/1","doc":null,"ref":"jsonpull_expect.html#float/1"},{"type":"function","title":"jsonpull_expect.integer/1","doc":null,"ref":"jsonpull_expect.html#integer/1"},{"type":"function","title":"jsonpull_expect.iolist/1","doc":null,"ref":"jsonpull_expect.html#iolist/1"},{"type":"function","title":"jsonpull_expect.key/1","doc":null,"ref":"jsonpull_expect.html#key/1"},{"type":"function","title":"jsonpull_expect.null/1","doc":null,"ref":"jsonpull_expect.html#null/1"},{"type":"function","title":"jsonpull_expect.number/1","doc":null,"ref":"jsonpull_expect.html#number/1"},{"type":"function","title":"jsonpull_expect.object/1","doc":null,"ref":"jsonpull_expect.html#object/1"},{"type":"function","title":"jsonpull_expect.string/1","doc":null,"ref":"jsonpull_expect.html#string/1"},{"type":"module","title":"jsonpull_read","doc":"These are the most basic functions you can use to read data from JSON. They take the value you ask for out of binary JSON if it's there. They return {Bin, Rest} or an atom representing an error. They will also skip whitespace. For more information on how to use these functions and what a pull parser is, read the Overview .","ref":"jsonpull_read.html"},{"type":"function","title":"jsonpull_read.array/1","doc":"Try to read an array from the JSON. If an array is starting here, the binary &lt;&lt;&quot;[&quot;&gt;&gt; will be returned. Structured types are handled differently in pull parsers. For more info, read the Overview.","ref":"jsonpull_read.html#array/1"},{"type":"function","title":"jsonpull_read.boolean/1","doc":"Try to read a boolean from the JSON. If it's there, it will be returned as a binary &lt;&lt;&quot;true&quot;&gt;&gt; or &lt;&lt;&quot;false&quot;&gt;&gt; .","ref":"jsonpull_read.html#boolean/1"},{"type":"function","title":"jsonpull_read.element/1","doc":"Try to read the next array element from the JSON. If there is an element, either the binary &lt;&lt;&quot;,&quot;&gt;&gt; or an empty binary will be returned. If the array is ending, the binary &lt;&lt;&quot;]&quot;&gt;&gt; will be returned. Structured types are handled differently in pull parsers. For more info, read the Overview.","ref":"jsonpull_read.html#element/1"},{"type":"function","title":"jsonpull_read.key/1","doc":"Try to read the next object key from the JSON. If there is a key, it will be returned as a raw string, with the same caveats as string/1 . If the object is ending, the binary &lt;&lt;&quot;}&quot;&gt;&gt; will be returned. Structured types are handled differently in pull parsers. For more info, read the Overview.","ref":"jsonpull_read.html#key/1"},{"type":"function","title":"jsonpull_read.null/1","doc":"Try to read a null from the JSON. If it's there, it will be returned as a binary &lt;&lt;&quot;null&quot;&gt;&gt; .","ref":"jsonpull_read.html#null/1"},{"type":"function","title":"jsonpull_read.number/1","doc":"Try to read a number from the JSON. If it's there, it will be returned as a binary string in its original form.","ref":"jsonpull_read.html#number/1"},{"type":"function","title":"jsonpull_read.object/1","doc":"Try to read an object from the JSON. If there is an object, the binary &lt;&lt;&quot;{&quot;&gt;&gt; will be returned. Structured types are handled differently in pull parsers. For more info, read the Overview.","ref":"jsonpull_read.html#object/1"},{"type":"function","title":"jsonpull_read.string/1","doc":"Try to read a string from the JSON. If it's there, it will be returned without any sort of processing. This means escape sequences and the quotation marks on the edges are still there. See also: jsonpull:unescape_and_strip/1 .","ref":"jsonpull_read.html#string/1"},{"type":"module","title":"jsonpull_construct","doc":"These functions are helpers to construct a proper Erlang list or map from JSON equivalents. They can be used with the main functions that return {ok, {Value, Rest}} or {error, Atom} and also have the same signature themselves. However, you can also substitute your own function to construct whatever you'd like.","ref":"jsonpull_construct.html"},{"type":"function","title":"jsonpull_construct.fold/3","doc":"Tries to read an object from the JSON and constructs a value using Conditions. The conditions can be seen as a list of items to check. Each condition will only apply once. It works like this: A new key is pulled from the object The key is checked to find any condition that matches it. If no condition is found, the value is skipped. When a key has a matching condition, the Fun for that condition is called to read the value. If it's a required condition, the constructor will make sure it only gets called once by erroring out if the condition matches any key again. If it's an optional condition, the constructor will make sure it only gets called once by silently ignoring any further mathces. When the object runs out of keys, if there are any required conditions left, an error will be returned. A single condition can have any number of Key-Fun pairs. Each pair contains the Key to check, and the Fun to be called if found. The Fun receives the JSON to read and the current accumulator. You may modify the accumulator in any way you'd like. There are some shortcut tuples you can use instead of supplying a Fun yourself, such as { set , name , string } % the third element can be a main-ish fun or an atom that refers to one instead of fun ( Bin , Acc ) -&gt; { ok , { Name , Rest } } = jsonpull : string ( Bin ) , { ok , { Acc #{ name =&gt; Name } , Rest } } end {set, Key, ...} will behave differently depending on the type of your accumulator. If it is a map, it will set the Key in the accumulator to the value. If it is a tuple, it will use setelement to set the corresponding Nth element to the value. If it is a list, it will be added to the list as {Key, Value} - like a proplist. This is still pretty complicated. Please look at some of the examples if you need more info.","ref":"jsonpull_construct.html#fold/3"},{"type":"function","title":"jsonpull_construct.list/2","doc":"Tries to read an array from the JSON and constructs a list using Fun. The Fun should have the same signature as one of the main reading functions. You can also use an atom as a shorthand for the main functions, e.g. string instead of 'fun jsonpull:string/1' .","ref":"jsonpull_construct.html#list/2"},{"type":"function","title":"jsonpull_construct.map/2","doc":"Equivalent to jsonpull_construct:fold(JSON, #{}, Conditions) .","ref":"jsonpull_construct.html#map/2"},{"type":"function","title":"jsonpull_construct.single_value/3","doc":"Helper function for pulling a single value out of an object. Equivalent to jsonpull_construct:fold(JSON, unreachable, [{required, [{Key, Fun}]}]) .","ref":"jsonpull_construct.html#single_value/3"},{"type":"extras","title":"A JSON pull parser for Erlang","doc":"# A JSON pull parser for Erlang\nI would like to start this message by referring you to the talk [A Fast, Compliant JSON Pull Parser for Writing Robust Applications](https://www.youtube.com/watch?v=_GrHKyUYyRc) by Jonathan Müller at CppCon 2023. Watching that talk is what inspired me to create this library and it will explain all of these concepts better than I can. I will try to sum it up briefly, but you can skip ahead if you did watch it.\n\nJSON parsing is something that's very familiar to anyone that has been programming for the web in the past decade. There already exists a number of libraries for this, so what makes this different? Well, it's a completely different paradigm.\n\nFirst of all, let's consider the most basic JSON parser: a DOM parser. It simply reads a JSON object and translates it, as best it can, into a native data type. However: We are not writing Javascript. We don't have objects and arrays, we have maps and lists. In some ways they are equivalent, in others, they are not. In many cases, you end up parsing your JSON data into a native type, and then \"parsing\" that native type into the format you actually want.\n\nThe second type of parser is a SAX parser. Instead of giving you the entire object, it sends a callback every time a value is read, leaving you to reconstruct a type from a chain of events. This sort of solves one problem, but it means you have to work with very difficult to understand syntax.\nFrom a quick search on Hex.pm, here is how I could categorize the current available libraries:\n- DOM parsers: jason, poison, jsx, yamerl, thoas, jsone, json\n- SAX parsers: jaxon\n\nSo what is a pull parser? It works like a lazy iterator available in many languages, or like Streams in Elixir (from what I understand, I'm not really an Elixir person). You only pull data when you want to, and what the pull parser lets you do is to manage your expectations. You get a few basic building blocks, and on top of these you can build complex constructions.\n\nHere are the JSON primitives:\n```erl\n{ok, {null, Rest}} = jsonpull:null( >).\n{ok, {true, Rest}} = jsonpull:boolean( >).\n% Strings may be pulled as binary strings or iolists\n{ok, { >, Rest}} = jsonpull:string( >).\n% Numbers may be pulled as the raw string interpretation or asserted to be an integer or float\n{ok, { >, Rest}} = jsonpull:number( >).\n{ok, {12345, Rest}} = jsonpull:integer( >).\n{ok, {123.45, Rest}} = jsonpull:float( >).\n```\nSimple enough, but what about structures? Let's start with an array.\n```erl\n{ok, {begin_array, R1}} = jsonpull:array( >).\n{ok, {element, R2}} = jsonpull:element(R1). % We use this to know there is an element to read.\n{ok, {1, R3}} = jsonpull:integer(R2).\n{ok, {element, R4}} = jsonpull:element(R3). % It also skips the comma and any whitespace.\n{ok, {2, R5}} = jsonpull:integer(R4).\n{ok, {element, R6}} = jsonpull:element(R5).\n{ok, {3, R7}} = jsonpull:integer(R6).\n{ok, {end_array,  >}} = jsonpull:element(R7).\n```\nSure, this is verbose for now, but we're now certain at every step of the way that we are reading 3 integers from an array in this JSON, and nothing else. We have confirmed the structure of our data in the parsing step and any invalid input is denied already.\nAnd now, on top of this we can already think of some constructs to make use of this iteration. For example, to read an array of numbers we can just do:\n```erl\n{ok, {[1,2,3],  >}} = jsonpull_construct:list( >, fun jsonpull:integer/1).\n```\nI'll talk about the different modules later. For now, let's proceed to the last type: objects. Objects are a difficult problem in any Erlang-adjacent JSON parser because you never know what you will get, or what you want. A proplist? A map? How about this: You decide.\n```erl\nJSON =  >,\n{ok, {begin_object, R1}} = jsonpull:object(JSON),\n{ok, { >, R2}} = jsonpull:key(R1),\nR3 = jsonpull:skip_value(R2),\n{ok, { >, R4}} = jsonpull:key(R3),\n{ok, {Name, R5}} = jsonpull:string(R4),\n{ok, { >, R6}} = jsonpull:key(R5),\n{ok, {IsSpeaker, R7}} = jsonpull:boolean(R6),\n{ok, {end_object,  >}} = jsonpull:key(R7),\n#{ name => Name, is_speaker => IsSpeaker }.\n```\nThis has one problem besides being clunky: Keys in JSON objects are not sorted. Therefore, we need to pull a key first and see what we're dealing with before we act on it. But sure enough, this is an easy problem to solve and there is another easy construct you can use:\n```erl\nJSON =  >,\njsonpull_construct:map(JSON, [\n    {required, [{ >, fun (Bin, Acc) ->\n        {ok, {N, Rest}} = jsonpull:string(Bin),\n        {ok, {Acc#{name => N}, Rest}}\n    end}]},\n    {optional, [{ >, fun (Bin, Acc) ->\n        {ok, {B, Rest}} = jsonpull:boolean(Bin),\n        {ok, {Acc#{is_speaker => B}, Rest}}\n    end}]}\n]).\n```\nAnd we can make that even better with some shortcuts:\n```erl\nJSON =  >,\njsonpull_construct:map(JSON, [\n\t{required, [{ >, {set, name, string}}]},\n\t{optional, [{ >, {set, is_speaker, boolean}}]}\n]).\n```\nAll of these example return the same map. But what if we could do something no other Erlang JSON parser could do?\n```erl\n-record(talk, {\n    name,\n    is_speaker\n}).\n\nJSON =  >,\n{ok, {Rec,  >}} = jsonpull_construct:fold(JSON, #talk{}, [\n\t{required, [{ >, {set, #talk.name, string}}]},\n\t{optional, [{ >, {set, #talk.is_speaker, boolean}}]}\n]),\n#talk{name =  >, is_speaker = true} = Rec.\n```\n---\nJSONPull is an experimental library and I have not yet completely nailed down the structure. I wouldn't recommend using it in a production setting unless you are prepared to change your code down the line in case of breaking changes.\n\nNevertheless, the different modules currently available in JSONPull follow:\n#","ref":"readme.html"},{"type":"extras","title":"jsonpull_read - A JSON pull parser for Erlang","doc":"This module is the completely barebones JSON reading tool that serves as the basis for the entire library. It will attempt to read exact types and return their binary data without modification. The common function signature here is `{ReadBinary, RestBinary} | ErrorAtom = jsonpull_read:TYPE(JSON)`\n#","ref":"readme.html#jsonpull_read"},{"type":"extras","title":"jsonpull - A JSON pull parser for Erlang","doc":"This module is the basic frontend. It reads and converts values to the type you want. The common function signature here is `{ok, {Value, RestBinary}} | {error, ErrorAtom} = jsonpull:TYPE(JSON)`\n#","ref":"readme.html#jsonpull"},{"type":"extras","title":"jsonpull_expect - A JSON pull parser for Erlang","doc":"In the pursuit of less boilerplate, I threw together this module. It is identical to the basic module except anything but ok will be thrown as an error, with a bit of error_info to match. The common function signature is `{Value, RestBinary} = jsonpull_expect:TYPE(JSON)`\n#","ref":"readme.html#jsonpull_expect"},{"type":"extras","title":"jsonpull_construct - A JSON pull parser for Erlang","doc":"This module holds the helpful constructs that make working with the structured types more enjoyable. For the elements of a list and the values of an object they call the user-supplied function to read JSON binary and come up with a type. This can either be your own custom fun, a pointer to a jsonpull fun, or just an atom which will be taken as `jsonpull:ATOM/1`.\n\nFor jsonpull_construct:list(JSON, Fun) any function that returns `{ok, {Value, RestBinary}}`will work.\n\njsonpull_construct:fold/3 and map/2 are slightly different. For you to make sure that the object was constructed successfully you will specify constraints that will have to be met for the entire function to exit correctly. Each constraint specifies if it is required or optional, and has a list of keys it is looking for and what to do with the value if found. The \"what to do\" can either be a function that works like in a fold or a helper like {set, key, TypeFun} where TypeFun follows the same rules as the list fun.\n\n---\nHere are some tidbits in closing:\n\nFunctions in this documentation are categorized by what base JSON type they pull from. For example, `jsonpull:null/1` obviously takes a JSON null while `jsonpull:existing_atom/1` exists in the String category.\n\nI think it would be nice to have a parser module in the future. It would hold the binary inside itself and enable you to write code like:\n```erl\ntrue = jsonpull_parser:boolean(P).\n```\n\nBecause of the architecture, the entire library is somewhat stream-compatible, although detection of when a value is cut off is lacking, namely with numbers.\n\nI have not benchmarked the library against other options, but I have tried my best to abide by Erlang's efficiency guide when it comes to binaries. It is written in 100% native Erlang with no NIFs.\n\nThe library is tested with unit tests, property tests (including fuzzing) and against the JSONTestSuite. You may view the JSONTestSuite results [here](assets/compat.html).\n\nCurrently, there is no JSON encoding in this library. I believe there is room for a similar paradigm when it comes to JSON encoding, but it is only a thought at the moment.","ref":"readme.html#jsonpull_construct"}],"content_type":"text/markdown"}